/*
 * vbi.c: w_pvrscan/wirbelscan - A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * wirbelscan-0.0.5: add support for vbi network names
 *
 * $Id$ v20080913
 */


#include <stdarg.h>
#include <stdlib.h>
#include "vbi.h"
#include "common.h"
#include "config.h"
#include <linux/videodev2.h>
#include <vdr/device.h>


#if HAVE_LIBZVBI == 1

#include <libzvbi.h>

static vbi_capture * vbiCapture;
static vbi_decoder * vbiDecoder;
unsigned int vbiServices = VBI_SLICED_TELETEXT_B | VBI_SLICED_VPS | VBI_SLICED_CAPTION_525;
const int vbiNumFrames = 11 * 30;
bool   vbiQuit;
char * vbiChannelNameLocal = NULL;


///!-----------------------------------------------------------------
///!  v 0.0.5, vbi event handler for zvbi
///!-----------------------------------------------------------------

static void vbiEventHandler(vbi_event * vbiEvent, void * Name) {
  const char *vbiChannelName = strdup((const char *)Name);
  
  switch (vbiEvent->type) {

    case VBI_EVENT_NETWORK:
    case VBI_EVENT_NETWORK_ID:
      if (vbiEvent->ev.network.name && strlen((const char *) vbiEvent->ev.network.name) > 0) {
        dlog(4, "********* %s ***********", (const char *) vbiEvent->ev.network.name);

        vbiChannelName = (const char *) vbiEvent->ev.network.name;

        vbiChannelNameLocal = vbi_strndup_iconv(vbi_locale_codeset(),
           "ISO-8859-1", vbiChannelName, strlen(vbiChannelName), '?');
        vbiQuit = true;
        }
      break;

    default:
      break;

    }
}

///!-----------------------------------------------------------------
///!  v 0.0.5, vbi scan loop for zvbi
///!-----------------------------------------------------------------

static void vbiLoop(void) {
  struct timeval vbiReadTimeout;
  vbi_capture_buffer * vbiCaptureBuffer;
  vbiReadTimeout.tv_sec  = 3;  // read timeout = 2sec
  vbiReadTimeout.tv_usec = 0;

  int startTime = time(NULL);
  int timeOut = 3;

  
  for (int frame = vbiNumFrames; frame > 0; --frame) {
    unsigned int vbiNumLines;
    int result = vbi_capture_pull_sliced(vbiCapture, &vbiCaptureBuffer, &vbiReadTimeout);
    switch(result) {
       case -1: dlog(0, "%s: vbi_capture_pull_sliced returned %d (%s)",
                     __FUNCTION__ , errno, strerror (errno));
                return;
       case 0:  dlog(5, "%s: vbi_capture_pull_sliced timeout.",__FUNCTION__);
                return;
       case 1:  break; // success
       default: dlog(0, "%s: vbi_capture_pull_sliced - unknown result %d",__FUNCTION__ , result);
                return;
       }
    vbiNumLines = vbiCaptureBuffer->size/sizeof(vbi_sliced);
    vbi_decode(vbiDecoder, (vbi_sliced *) vbiCaptureBuffer->data, vbiNumLines, vbiCaptureBuffer->timestamp);
    if (vbiQuit || (time(NULL) > startTime + timeOut)) return;
    }
}

///!-----------------------------------------------------------------
///!  v 0.0.5, get channel network name from vbi.
///! returns network name if found, channelName otherwise
///! BE CAREFUL HERE: DO NOT START TWO INSTANCES, RESULT WILL BE
///! UNPREDICTABLE. --wirbel 
///!-----------------------------------------------------------------

cString vbiScanChannelName(int device, const cString channelName) {

  cString vbiDevice   = cString::sprintf("/dev/vbi%d", device);
  cString videoDevice = cString::sprintf("/dev/video%d", device);
  cString Name        = cString::sprintf("%s", *channelName);
  char * vbiErr;
  int err = 0;  

  struct v4l2_format       vbifmt;
  struct v4l2_ext_control  ctrl;
  struct v4l2_ext_controls ctrls = { V4L2_CTRL_CLASS_MPEG, 1, 0, {0,0}, &ctrl };
  struct v4l2_encoder_cmd  eCmd;

  memset(&vbifmt, 0, sizeof(vbifmt));
  memset(&eCmd,   0, sizeof(v4l2_encoder_cmd));
  vbiQuit = false;
  vbiChannelNameLocal = NULL;

  int video_fd = open(*videoDevice, O_RDWR); 

  eCmd.cmd = V4L2_ENC_CMD_STOP;  
  if ((err = ioctl(video_fd, VIDIOC_ENCODER_CMD, &eCmd))) {
    dlog(0, "cannot stop encoder :("); 
    }
  if (!err) {
    int vbi_fd;
    ctrl.id    = V4L2_CID_MPEG_STREAM_VBI_FMT;
    ctrl.value = V4L2_MPEG_STREAM_VBI_FMT_IVTV;
    vbifmt.fmt.sliced.service_set = V4L2_SLICED_TELETEXT_B;
    vbifmt.type = V4L2_BUF_TYPE_SLICED_VBI_CAPTURE;
    vbifmt.fmt.sliced.reserved[0] = 0;
    vbifmt.fmt.sliced.reserved[1] = 0;
    vbi_fd = open(*vbiDevice, O_RDWR);

    if (err = ioctl(vbi_fd, VIDIOC_S_EXT_CTRLS, &ctrls)) {
      dlog(0, "%s %d: error setting vbi mode, %d:%s", __FUNCTION__, __LINE__, errno, strerror(errno));
      }
    if (err = ioctl(vbi_fd, VIDIOC_S_FMT, &vbifmt)) {
      dlog(0, "%s %d: error setting vbi fmt, __FUNCTION__, %d:%s", __LINE__, errno, strerror(errno));
      }
    close(vbi_fd);
    }
  if (! err) {
    memset(&eCmd, 0, sizeof(v4l2_encoder_cmd));
    eCmd.cmd = V4L2_ENC_CMD_START;  
    if ((err = ioctl(video_fd, VIDIOC_ENCODER_CMD, &eCmd))) {
      dlog(0, "cannot start encoder :("); 
      }
    }
  close(video_fd);



  vbiCapture = vbi_capture_v4l2_new (*vbiDevice, 5, &vbiServices, 0, &vbiErr, FALSE);

  if (! vbiCapture) {
    dlog(0, "    Error: Cannot capture VBI: %s", vbiErr);
    free (vbiErr);
    return Name;
    }

  vbiDecoder = vbi_decoder_new();

  if (! vbiDecoder) {
    dlog(0, "    Error: Cannot create VBI decoder");
    return Name;
    }

  vbi_channel_switched(vbiDecoder, 0);  

  if (! vbi_event_handler_add(vbiDecoder, VBI_EVENT_NETWORK | VBI_EVENT_NETWORK_ID, vbiEventHandler, (char *) *Name)) {
    dlog(0, "    Error: Could not add VBI event handler.");
    return Name;
    }

  vbiLoop();

  if (vbiChannelNameLocal) {
    Name = strdup(vbiChannelNameLocal);
    free (vbiChannelNameLocal);
    vbiChannelNameLocal=NULL;
    }

  vbi_channel_switched(vbiDecoder, 0); 
  vbi_decoder_delete (vbiDecoder);
  vbiDecoder = NULL;
  vbi_capture_delete (vbiCapture);
  vbiCapture = NULL;
  return Name;
}

#else

///!-----------------------------------------------------------------
///!  v 0.0.5, dummy for non-zvbi compiling.
///!-----------------------------------------------------------------
# warning : ***** NO LIBZVBI WAS FOUND - DISABLING VBI SCAN FOR ANALOG CHANNELNAMES  *****

cString vbiScanChannelName(int device, const cString channelName) {
  dlog(3, "    %s (no zvbi support)", *channelName);
  return channelName;
}  

#endif

